'''
MIT License

Copyright (c) 2024 Saurabh Shivpuje

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import numpy as np
from itertools import combinations, product
from scipy.linalg import expm
from numpy import kron

# Ansatz Class and Operators

class Ansatz_class:
    def __init__(self, nqbit, u0, relrcut, pool, theta=[], ansatz=[]):
        self.theta = np.array(theta)
        self.A = ansatz
        self.state = u0.copy()
        self.ref = u0.copy()
        self.relrcut = relrcut
        self.nqbit = nqbit
        self.pool = pool

class AnsatzOperatorBase:
    pass

class PauliOperator_class(AnsatzOperatorBase):
    def __init__(self, mat, tag, nqbit):
        self.mat = mat
        self.tag = tag
        self.nqbit = nqbit

def single_clause(ops, q_ind, weight, num_qubit):
    si = np.eye(2)
    res = weight * np.eye(1)
    for i in range(1, num_qubit + 1):
        if i in q_ind:
            op2 = eval(f"{ops[q_ind.index(i)]}")
            res = kron(res, op2)
        else:
            res = kron(res, si)
    return res

def PauliOperator(ops, idx, w, nqbit):
    sortedRep = sorted(zip(ops, idx), key=lambda x: x[1])
    tag = ''.join([f"{o}{i}" for o, i in sortedRep])
    mat = single_clause(ops, idx, w, nqbit)
    return PauliOperator_class(mat, tag, nqbit)

# Defining pool of operators/gates
sx = np.array([[0, 1], [1, 0]])
sy = np.array([[0, -1j], [1j, 0]])
sz = np.array([[1, 0], [0, -1]])
# S = np.array([[1, 0], [0, 1j]])

def build_pool(nqbit):
    pauliStr = ["sx", "sz", "sy"]
    res = []
    for order in range(1, nqbit+1):
        for idx in combinations(range(1, nqbit + 1), order):
            for op in product(pauliStr, repeat=order):
                res.append(PauliOperator(op, list(idx), 1, nqbit))
    return res

def Ansatz(u0, relrcut, theta=None, ansatz=None, vectorized=False):

    if theta is None: theta = []
    if ansatz is None: ansatz = []
    nqbit = int(np.log2(len(u0)))

    if vectorized:
        pool_qubit =  2*nqbit
        u0 = np.outer(u0, np.conjugate(u0)).flatten()
    else:
        pool_qubit = nqbit

    # Build operator pool
    pool = build_pool(pool_qubit)
    return Ansatz_class(nqbit, u0, relrcut, pool, theta, ansatz)


import numpy as np
from qiskit import QuantumRegister, QuantumCircuit

from .params import I,X,Y,Z

##=======the Walsh operator scheme===========
#1. functions for Walsh representation
#(a). the binary and qubit state representation of a integer
#(b). the function to calculate walsh coefficient
#this is binary representation of a integer j!
#j=[j1,j2,j3] for example: 3=[1,1,0]
def binary(j,nbits):
  arr = np.zeros(nbits,dtype=int)
  res = j
  for i in range(nbits):
    if(res>1):
        arr[i] = res%2
        res=res//2
    else:
        arr[i] = res
        break
  return arr

#input: array of 0 or 1
#output: max index of 1
#e.g.: j=4 binary is [0,0,1,0], max_bit is 2
def max_bit(arr):
  iflag = 0
  for i in range(len(arr)-1,-1,-1):
    if(arr[i]==1):
        iflag = i
        break
  return iflag

#j-th gray code:
#first do the right shift of the binary representation
#then do the bit wise XOR
def gray(j,nbits):
  arr_n1 = binary(j//2,nbits) #
  arr_j = binary(j,nbits) #
  arr_res = np.zeros_like(arr_n1)
  for i in range(nbits):
    if(arr_n1[i]==arr_j[i]):
        arr_res[i] = 0
    else:
        arr_res[i] = 1
  return arr_res

# return the decimal of arr_j
# [1,1,0] is j0=1,j1=1, which is 3
def decimal(arr_j):
  res = 0
  nbits = len(arr_j)
  for i in range(nbits):
    res+= 2**i*arr_j[i]
  return res

#2. The Walsh coefficient and operator
#input an fk array
#(that's the discretize array in qubit statevector) |qn,...,q1>
#e.g. f3 is the coefficient of |0,1,1>, which is q0=1, q1=1
#this just match the definition of function binary
#output the coefficient in the walsh function representation
def walsh_coef(fk,nbits):
  nlen = 2**nbits
  if(nlen!=len(fk)): print('walsh_coef: err in array length')

  a_arr = np.zeros(nlen)
  for i in range(nlen):
    arr_j = binary(i,nbits)
    for k in range(nlen):
      fac = (-1)**(np.dot(arr_j,binary(k,nbits)))
      a_arr[i] += fk[k]*fac
    a_arr[i] /= nlen
  return a_arr

#walsh operator, \Prod \otimes Z^(j)
#j=[1,1,0] return IZZ (q0,q1 is Z, q2 is I, in qiskit convention order)
def walsh_oprQ(arr_j,nbits):
  res = 1.0
  for i in range(nbits-1,-1,-1):
    if(arr_j[i]==0):
      res = np.kron(res,I)
    elif(arr_j[i]==1):
      res = np.kron(res,Z)
    else:
      print('err')
  return res


#3. Walsh circuit list and Gray code optimize for diagonal unitaries
#  functions to generate the quantum circuit of the diagnoal unitaries in the Walsh Operator representation
#input an walsh coefficient (arra) of f_k array,
#output the list that contain all the parameters to construct the circuit for U=diag(e^(i f_k))
def cirq_list_walsh(arra,Nqb,epsilon):
  if(len(arra)!=2**Nqb): print('err')

  U=[]

  for j in range(1,2**Nqb):

    #encoding: Gray Code or Binary Code
    #arr_j = binary(j,Nqb)
    arr_j = gray(j,Nqb)
    j_code = decimal(gray(j,Nqb))

    if(abs(arra[j_code])<epsilon): continue

    max_index = max_bit(arr_j) #the R gate will located at the max index

    for ibit in range(max_index):
      if(arr_j[ibit]==1):
        U.append(('C',ibit,max_index))

    U.append(('R',max_index,-2.0*arra[j_code]))

    for ibit in range(max_index-1,-1,-1):
      if(arr_j[ibit]==1):
        U.append(('C',ibit,max_index))
  return U

#generate circuit from list of turple U
def cirq_from_U(U,Nqb):
  nlen = len(U)

  #the quantum circuit
  qr = QuantumRegister(Nqb)
  #cr = ClassicalRegister(1)
  qc = QuantumCircuit(qr)

  for i in range(nlen):
    gate_str = U[i]
    if(gate_str[0]=='C'):
      qc.cx(gate_str[1],gate_str[2])
    elif(gate_str[0]=='R'):
      qc.rz(gate_str[2],gate_str[1])
    else:
      print('err in reading circuit')
  return qc


#Optimize the quantum circuit
def optimize(U):
  U0,iflag0 = scanI(U)
  U0,iflag1 = scanRule2(U0)

  #some possible exchange and check
  i = 0
  while(i<len(U0)-1):

    if(U0[i+1][0] == 'C' and U0[i][0] =='C'):

      ictrl = U0[i][1]; itarg = U0[i][2]
      if(ictrl == U0[i+1][1] or itarg == U0[i+1][2]):
        Utmp = U0.copy()
        Utmp[i] = U0[i+1]
        Utmp[i+1] = U0[i]
        Utmp,iflag0 = scanI(Utmp)
        Utmp,iflag1 = scanRule2(Utmp)

        #if(iflag0 or iflag1): U0 = Utmp
        U0 = Utmp

    i += 1
  return U0

#scan if there are identical adjacent CNOT gate
def scanI(U):
  nlen = len(U)
  newU = []
  i=0
  iflag = 0
  while(i<nlen-1):
    if(U[i]==U[i+1] and U[i][0] =='C'):
      i += 2
      iflag = 1
    else:
      newU.append(U[i])
      i += 1
  #end
  if(i==nlen-1): newU.append(U[nlen-1])
  return newU,iflag

#scan if there have 3 gate can reduced to:
#the target of one CNOT is the control of another.
#CijCjk = CjkCikCij
def scanRule2(U):
  nlen = len(U)
  newU = []
  i=0
  iflag = 0
  while(i<nlen-2):

    if(U[i][0] == U[i+1][0] == U[i+2][0] =='C'):
      sj = U[i][1]
      sk = U[i][2]
      si = U[i+1][1]
      if(sj == U[i+2][2] and si == U[i+2][1] and sk == U[i+1][2]):
        i += 3
        newU.append(('C',si,sj))
        newU.append(('C',sj,sk))
        print('Rule2')
        iflag = 1
      else:
        newU.append(U[i])
        i += 1
    else:
      newU.append(U[i])
      i += 1
  #end
  if(i==nlen-2):
    newU.append(U[nlen-2])
    newU.append(U[nlen-1])
  return newU,iflag
##=======end the Walsh operator scheme===========

##=======test functions for Walsh operator scheme=========
#extented matrix of the quantum gate
#note that this is the reverse order with qiskit convention
def ext_cx(ictrl,itarg,Nqb):
  if(ictrl>=Nqb or itarg>=Nqb): print('err')
  Proj00 = np.array([[1.0,0],[0,0]])
  Proj11 = np.array([[0.0,0],[0,1]])

  res00 = 1
  res11 = 1
  for i in range(Nqb-1,-1,-1):
  #for i in range(Nqb):
    if(i!=ictrl and i!=itarg):
      res00 = np.kron(res00,I)
      res11 = np.kron(res11,I)
    if(i==ictrl):
      res00 = np.kron(res00,Proj00)
      res11 = np.kron(res11,Proj11)
    if(i==itarg):
      res00 = np.kron(res00,I)
      res11 = np.kron(res11,X)
  return res00+res11

def ext_Rz(phi,iqb,Nqb):
  if(iqb>=Nqb): print('err')
  mat_rz = np.array([[np.exp(-1j*phi/2),0],[0,np.exp(1j*phi/2)]])
  res = 1
  for i in range(Nqb-1,-1,-1):
  #for i in range(Nqb):
    if(i!=iqb):
      res = np.kron(res,I)
    if(i==iqb):
      res = np.kron(res,mat_rz)
  return res

def ext_Z(iqb,Nqb):
  if(iqb>=Nqb): print('err')
  res = 1
  for i in range(Nqb-1,-1,-1):
  #for i in range(Nqb):
    if(i!=iqb):
      res = np.kron(res,I)
    if(i==iqb):
      res = np.kron(res,Z)
  return res

#generate matrix of quantum circuit from list of turple U
#convient for verify the result
def cirqmat_from_U(U,Nqb):
  nlen = len(U)

  Umat = []

  for i in range(nlen):
    gate_str = U[i]
    if(gate_str[0]=='C'):
      Umat.append(ext_cx(gate_str[1],gate_str[2],Nqb))
    elif(gate_str[0]=='R'):
      Umat.append(ext_Rz(gate_str[2],gate_str[1],Nqb))
    else:
      print('err in reading circuit')

  #connect all the gates
  res = Umat[-1]
  for i in range(nlen-2,-1,-1):
    res = res@Umat[i]
  return res

#input an walsh coefficient (arra) of f_k array, output the circuit matrix for U=diag(e^(i f_k))
def cirqmat_walsh(arra,Nqb,epsilon):
  if(len(arra)!=2**Nqb): print('err')

  U = []

  for j in range(1,2**Nqb):
    if(abs(arra[j])<epsilon): continue
    arr_j = binary(j,Nqb)
    max_index = max_bit(arr_j) #the R gate will located at the max index

    for ibit in range(max_index):
      if(arr_j[ibit]==1):
        U.append(ext_cx(ibit,max_index,Nqb))

    U.append(ext_Rz(-2*arra[j],max_index,Nqb))

    for ibit in range(max_index-1,-1,-1):
      if(arr_j[ibit]==1):
        U.append(ext_cx(ibit,max_index,Nqb))

  #connect all the gates
  res = U[-1]
  for i in range(len(U)-2,-1,-1):
    res = res@U[i]
  return res
#============================


import numpy as np
import numpy.typing as npt
from typing import List, Optional, Tuple

from qiskit import QuantumCircuit
from qiskit_aer.primitives import EstimatorV2 as Estimator
from qiskit.quantum_info import SparsePauliOp
from qiskit_aer.noise import NoiseModel
from qiskit_ibm_runtime.fake_provider import FakeSherbrooke

# To change the ansatz, apply_param and measure_der must both be modified.
def apply_param(
    params: npt.NDArray[np.float64], i: int, qc: QuantumCircuit, N: int
) -> None:
    """Apply parameter i to the quantum circuit currently constructing the ansatz.
        The ansatz must be built in a peicewise manner to allow for hadamard tests
        of the generators of the parameters to later be inserted to measure the A_ij
        and C_i matrix elements.

    Args:
        params (numpy.array): An array containing the values of all the ansatz parameters.
        parameter (int): Index of the parameter being applied.
        qc (QuantumCircuit): The qiskit ansatz quantum circuit currently being constructed.
        N (int): Number of qubits
    """
    qc.rx(params[i], i % N)
    if i % N == N - 1 and i != len(params) - 1:
        for i in range(N - 1):
            qc.cz(i, i + 1)


def measure_der(i: int, qc: QuantumCircuit, N: int) -> None:
    """Append a Hadamard test to the circuit to measure the generator of parameter i in the ansatz.
        The ansatz currently used is simply the two-local ansatz with only rx gates.
        Therefore the generator is only x-gates on the corresponding qubit.

    Args:
        parameter (int): The index of the parameter whose generator will be measured.
        qc (QuantumCircuit): The qiskit quantum circuit which is in the process of assembling the ansatz.
        N (int): Number of qubits
    """
    qc.cx(N, i % N)


def A_Circuit(params: npt.NDArray[np.float64], i: int, j: int, N: int) -> QuantumCircuit:
    """Constructs the qiskit quantum circuits used to measure each element of the A_ij matrix.

    Args:
        params (numpy.array): A numpy array containing the values of each parameter of the ansatz.
        i (int): The index from A_ij.  This also corresponds to the ansatz parameter i being measured.
        j (int): The index from A_ij.  This also corresponds to the ansatz parameter j being measured.
        N (int): The number of qubits.

    Returns:
        QuantumCircuit: The quantum circuit for an element of the A_ij matrix, in the form of a
            hadamard test of the generators of parameters i and j.  The value of the A_ij matrix
            can be found by measuring the ancilla qubit (qubit N) in the Z basis.
    """
    qc = QuantumCircuit(N + 1, 1)
    qc.h(N)
    for parameter in range(len(params)):  # Apply parameterized gates
        if parameter == i:
            qc.x(N)
            measure_der(parameter, qc, N)  # Measure generator for i
            qc.x(N)
        if parameter == j:
            measure_der(parameter, qc, N)  # Measure second generator for j
        apply_param(params, parameter, qc, N)
    qc.h(N)
    return qc


def Measure_A(
    init_circ: QuantumCircuit,
    params: npt.NDArray[np.float64],
    N: int,
    shots: int = 2**10,
    noisy: bool = False,
) -> npt.NDArray[np.float64]:
    """Create the A_ij matrix through measuring quantum circuits corresponding to each element.

    Args:
        init_circ (QuantumCircuit): The qiskit circuit representing the initial state of the system.
        params (numpy.array): A numpy array which contains the values of each parameter of the ansatz.
        N (int): The number of qubits.
        shots (int, optional): The number of shots used to estimate each element of the A_ij matrix. Defaults to 2**10.
        noisy (bool, optional): A boolean used to turn on and off the Fake-Sherbrooke qiskit noisy backend. Defaults to False.

    Returns:
        numpy.array: The A_ij matrix
    """
    A = [[0.0 for i in range(len(params))] for j in range(len(params))]
    for i in range(len(params)):
        for j in range(len(params) - i):
            qc = QuantumCircuit(N + 1, 1)
            ansatz = A_Circuit(params, i, i + j, N)
            qc = qc.compose(init_circ, [k for k in range(N)])
            qc = qc.compose(ansatz, [k for k in range(N + 1)])

            observable = SparsePauliOp.from_list([("Z" + "I" * N, 1.0)])
            if noisy:
                device_backend = FakeSherbrooke()
                coupling_map = device_backend.coupling_map
                noise_model = NoiseModel.from_backend(device_backend)
                basis_gates = noise_model.basis_gates
                estimator = Estimator(options={
                                    "backend_options":{"noise_model": noise_model},
                                    "run_options":{"shots": shots}}
                )
            else:
                estimator = Estimator(options={"run_options":{"shots": shots}})
            result = estimator.run([(qc, observable)]).result()
            A[i][i + j] = 0.25 * result[0].data.evs
    return np.array(A)


def C_Circuit(
    params: npt.NDArray[np.float64],
    i: int,
    pauli_string: str,
    N: int,
    evolution_type: str = "real",
) -> QuantumCircuit:

    """Create the qiskit quantum circuits to measure each element of the C_i vector.

    Args:
        params (numpy.array): A numpy array which contains the values of each parameter of the ansatz.
        i (int): The index of the C_i vector being measured. This also corresponds
            to the index i of the parameter whose generator will be measured
        pauli_string (str): A string containing a description of the pauli operator of the Hamiltonian which will be measured.
        N (int): The number of qubits.
        evolution_type (str, optional): This determines if the evolution will be real-time or imaginary-time
            through the addition of an extra gate. Defaults to "real".

    Returns:
        QuantumCircuit: The quantum circuit for an element of the C_i matrix, in the form of a
            hadamard test of the generators of parameter i.  The value of the C_i matrix
            can be found by measuring the ancilla qubit (qubit N) in the Z basis.
    """
    qc = QuantumCircuit(N + 1, 1)
    qc.h(N)
    if evolution_type == "imaginary":
        qc.s(N)  # To get only imaginary component
    else:
        qc.z(N)
    for parameter in range(len(params)):  # Apply parameterized gates
        if parameter == i:
            qc.x(N)
            measure_der(parameter, qc, N)  # Measure generators
            qc.x(N)
        apply_param(params, parameter, qc, N)
    pauli_measure(qc, pauli_string)
    qc.h(N)
    return qc


def Measure_C(
    init_circ: QuantumCircuit,
    params: npt.NDArray[np.float64],
    H: SparsePauliOp,
    N: int,
    shots: int = 2**10,
    evolution_type: str = "real",
    noisy: bool = False,
) -> npt.NDArray[np.float64]:
    """Create the C_i vector through measuring quantum circuits corresponding to each element.

    Args:
        init_circ (QuantumCircuit): A qiskit circuit constructing the initial state of the system.
        params (numpy.array): A numpy array containing the values of the parameters of the ansatz.
        H (SparsePauliOp): The Hamiltonian.
        N (int): The number of qubits.
        shots (int, optional): The number of shots to be used to measure each element of the C_i vector. Defaults to 2**10.
        evolution_type (str, optional): This determines if the evolution will be real-time or imaginary-time
            through the addition of an extra gate. Defaults to "real".
        noisy (bool, optional): A boolean used to turn on and off the Fake-Sherbrooke qiskit noisy backend. Defaults to False.

    Returns:
        numpy.array: The C_i vector.
    """
    C = [0.0 for i in range(len(params))]
    for i in range(len(params)):
        for pauli_string in range(len(H.paulis)):
            qc = QuantumCircuit(N + 1, 1)
            ansatz = C_Circuit(
                params, i, H.paulis[pauli_string], N, evolution_type=evolution_type
            )
            qc = qc.compose(init_circ, [k for k in range(N)])
            qc = qc.compose(ansatz, [k for k in range(N + 1)])
            observable = SparsePauliOp.from_list([("Z" + "I" * N, 1.0)])
            if noisy:
                device_backend = FakeSherbrooke()
                coupling_map = device_backend.coupling_map
                noise_model = NoiseModel.from_backend(device_backend)
                basis_gates = noise_model.basis_gates
                estimator = Estimator(options={
                                        "backend_options":{"noise_model": noise_model},
                                        "run_options":{"shots": shots}}
                    )
            else:
                estimator = Estimator(options={"run_options":{"shots": shots}})
            result = estimator.run([(qc, observable)]).result()

            C[i] -= 1 / 2 * H.coeffs[pauli_string].real * result[0].data.evs
    return np.array(C)


def pauli_measure(qc: QuantumCircuit, pauli_string: str) -> None:
    """Measure the given pauli string on the provided quantum circuit using a hadamard test.

    Args:
        qc (QuantumCircuit): The quantum circuit ansatz being constructed.
        pauli_string (str): The pauli string to be measured as a string.
    """
    N = len(pauli_string)
    for i in range(len(pauli_string)):  # Measure Pauli Strings
        if str(pauli_string[i]) == "X":
            qc.cx(N, i)
        if str(pauli_string[i]) == "Y":
            qc.cy(N, i)
        if str(pauli_string[i]) == "Z":
            qc.cz(N, i)


def Construct_Ansatz(
    init_circ: QuantumCircuit, params: npt.NDArray[np.float64], N: int
) -> QuantumCircuit:
    """Construct the full ansatz for use in measuring observables.

    Args:
        init_circ (QuantumCircuit): A qiskit circuit constructing the initial state of the system.
        params (numpy.array): A numpy vector containing the values of the parameters of the ansatz at a specific time.
        N (int): The number of qubits.

    Returns:
        QuantumCircuit: The full ansatz as a qiskit.QuantumCircuit.
    """
    qc = QuantumCircuit(N, 0)
    qc = qc.compose(init_circ, [k for k in range(N)])

    ansatz = QuantumCircuit(N, 0)
    for parameter in range(len(params)):  # Apply parameterized gates
        apply_param(params, parameter, ansatz, N)

    qc = qc.compose(ansatz, [k for k in range(N)])
    return qc


def ansatz_energy(
    init_circ: QuantumCircuit,
    params: npt.NDArray[np.float64],
    H: SparsePauliOp,
    shots: int = 2**14,
    noisy: bool = False,
) -> Tuple[float, float]:
    """Measure the energy of the ansatz.

    Args:
        init_circ (QuantumCircuit): A qiskit circuit constructing the initial state of the system.
        params (numpy.array): A numpy vector containing the values of the parameters of the ansatz at a specific time.
        H (SparsePauliOp): The Hamiltonian.
        shots (_type_, optional): The number of shots to be used to measure the energy. Defaults to 2**14.
        noisy (bool, optional): A boolean used to turn on and off the Fake-Sherbrooke qiskit noisy backend. Defaults to False.

    Returns:
        (float, float): Return (energy, variance) from the measured observables.
    """
    N = H.num_qubits

    if noisy:
        device_backend = FakeSherbrooke()
        coupling_map = device_backend.coupling_map
        noise_model = NoiseModel.from_backend(device_backend)
        basis_gates = noise_model.basis_gates
        estimator = Estimator(options={
                                "backend_options":{"noise_model": noise_model},
                                "run_options":{"shots": shots}}
                                )
    else:
        estimator = Estimator(options={"run_options":{"shots": shots}})
    qc = Construct_Ansatz(init_circ, params, N)
    result = estimator.run([(qc, H)]).result()
    return result[0].data.evs, result[0].data.stds


def VarQRTE(
    n_reps_ansatz: int,
    hamiltonian: SparsePauliOp,
    total_time: float = 1.0,
    timestep: float = 0.1,
    init_circ: Optional[QuantumCircuit] = None,
    shots: int = 2**10,
    noisy: bool = False,
) -> List[npt.NDArray[np.float64]]:
    """The Variational Quantum Real Time Evolution (VarQRTE) algorithm.  This uses quantum circuits to measure
        the elements of two objects, the A_ij matrix and the C_i vector.

    Args:
        n_reps_ansatz (int): The number of repetitions of the variational ansatz used to simulate Real-Time evolution.
        hamiltonian (SparsePauliOp): The Hamiltonian of the system.
        total_time (float, optional): A float to determine the total evolution time of the quantum system. Defaults to 1.0.
        timestep (float, optional): A float to determine the size of a single timestep. Defaults to 0.1.
        init_circ (QuantumCircuit, optional): A qiskit circuit constructing the initial state of the system.. Defaults to None.
        shots (int, optional): Number of shots to be used to measure observables. Defaults to 2**10.
        noisy (bool, optional): A boolean used to turn on and off the Fake-Sherbrooke qiskit noisy backend. Defaults to False.

    Returns:
        numpy.array: An array containing all the parameter values of the ansatz throughout its time evolution.
            These values can be put into Construct_Ansatz, or anstaz_energy to obtain observables of the system.
    """
    if init_circ is None:
        init_circ = QuantumCircuit(hamiltonian.num_qubits)

    initial_params = np.zeros(hamiltonian.num_qubits * (n_reps_ansatz + 1))
    num_timesteps = int(total_time / timestep)
    all_params = [np.copy(initial_params)]
    my_params = np.copy(initial_params)  # Reset Initial Parameters after each run
    for i in range(num_timesteps):
        print(f"Simulating Time={str(timestep*(i+1))}                      ", end="\r")
        theta_dot = np.array([0.0 for j in range(len(my_params))])
        A = Measure_A(
            init_circ, my_params, hamiltonian.num_qubits, shots=shots, noisy=noisy
        )
        C = Measure_C(
            init_circ,
            my_params,
            hamiltonian,
            hamiltonian.num_qubits,
            shots=shots,
            evolution_type="real",
            noisy=noisy,
        )

        # Approximately invert A using Truncated SVD
        u, s, v = np.linalg.svd(A)
        for j in range(len(s)):
            if s[j] < 1e-2:
                s[j] = 1e8
        t = np.diag(s**-1)
        A_inv = np.dot(v.transpose(), np.dot(t, u.transpose()))

        theta_dot = np.matmul(A_inv, C)

        my_params -= theta_dot * timestep
        all_params.append(np.copy(my_params))
    return all_params


def VarQITE(
    n_reps_ansatz: int,
    hamiltonian: SparsePauliOp,
    total_time: float,
    timestep: float,
    init_circ: Optional[QuantumCircuit] = None,
    shots: int = 2**10,
    noisy: bool = False,
) -> List[npt.NDArray[np.float64]]:
    """The Variational Quantum Imaginary Time Evolution (VarQITE) algorithm.  This uses quantum circuits to measure
        the elements of two objects, the A_ij matrix and the C_i vector.

    Args:
        n_reps_ansatz (int): The number of repetitions of the variational ansatz used to simulate Imaginary-Time evolution.
        hamiltonian (SparsePauliOp): The Hamiltonian of the system.
        total_time (float, optional): A float to determine the total evolution time of the quantum system. Defaults to 1.0.
        timestep (float, optional): A float to determine the size of a single timestep. Defaults to 0.1.
        init_circ (QuantumCircuit, optional): A qiskit circuit constructing the initial state of the system.. Defaults to None.
        shots (int, optional): Number of shots to be used to measure observables. Defaults to 2**10.
        noisy (bool, optional): A boolean used to turn on and off the Fake-Sherbrooke qiskit noisy backend. Defaults to False.

    Returns:
        numpy.array: An array containing all the parameter values of the ansatz throughout its time evolution.
            These values can be put into Construct_Ansatz, or anstaz_energy to obtain observables of the system.
    """
    if init_circ is None:
        init_circ = QuantumCircuit(hamiltonian.num_qubits)

    initial_params = np.zeros(hamiltonian.num_qubits * (n_reps_ansatz + 1))
    num_timesteps = int(total_time / timestep)
    all_params = [np.copy(initial_params)]

    my_params = np.copy(initial_params)  # Reset Initial Parameters after each run
    for i in range(num_timesteps):
        print(f"Timestep: {str(i*timestep)}                      ", end="\r")
        theta_dot = np.array([0.0 for j in range(len(my_params))])
        A = np.array(
            Measure_A(
                init_circ, my_params, hamiltonian.num_qubits, shots=shots, noisy=noisy
            )
        )
        C = np.array(
            Measure_C(
                init_circ,
                my_params,
                hamiltonian,
                hamiltonian.num_qubits,
                shots=shots,
                noisy=noisy,
                evolution_type="imaginary",
            )
        )

        # Approximately invert A using Truncated SVD
        u, s, v = np.linalg.svd(A)
        for j in range(len(s)):
            if s[j] < 1e-2:
                s[j] = 1e7
        t = np.diag(s**-1)
        A_inv = np.dot(v.transpose(), np.dot(t, u.transpose()))
        # A_inv=np.dot(v,np.dot(t,u.transpose()))

        theta_dot = np.matmul(A_inv, C)

        my_params += theta_dot * timestep
        all_params.append(np.copy(my_params))

        # print("Theta dot: "+str(np.sum(np.abs(theta_dot))))
        # print("(Energy ,Variance): "+str(ansatz_energy(init_circ, my_params[:], hamiltonian)))
        # print()
    return all_params



from qiskit.quantum_info import SparsePauliOp


# -----------------------------------------------------------------
# Hamiltonian Functions
# -----------------------------------------------------------------
def get_hamiltonian_n_site_terms(n, coeff, n_qubits):
    '''
        Assembles each term in the Hamiltonian based on their Pauli string
        representation and multiplying by the respective coefficient.
    '''
    XX_coeff = coeff[0]
    YY_coeff = coeff[1]
    ZZ_coeff = coeff[2]
    Z_coeff = coeff[3]

    XX_term = SparsePauliOp(("I" * n + "XX" + "I" * (n_qubits - 2 - n)))
    XX_term *= XX_coeff
    YY_term = SparsePauliOp(("I" * n + "YY" + "I" * (n_qubits - 2 - n)))
    YY_term *= YY_coeff
    ZZ_term = SparsePauliOp(("I" * n + "ZZ" + "I" * (n_qubits - 2 - n)))
    ZZ_term *= ZZ_coeff
    Z_term = SparsePauliOp(("I" * n + "Z" + "I" * (n_qubits - 1 - n)))
    Z_term *= Z_coeff

    return (XX_term + YY_term + ZZ_term + Z_term)


def get_heisenberg_hamiltonian(n_qubits, coeff=None):
    r'''
        Takes an integer number corresponding to number of spins/qubits
        and a list of sublists containing the necessary coefficients
        to assemble the complete Hamiltonian:
        $$
            H = \sum _i ^N h_z Z_i
                + \sum _i ^{N-1} (h_xx X_iX_{i+1}
                                  + h_yy Y_iY_{i+1}
                                  + h_zz Z_iZ_{i+1}
                                 )
        $$
        Each sublist contains the [XX, YY, ZZ, Z] coefficients in this order.
        The last sublist should have the same shape, but only the Z component
        is used.
        If no coefficient list is provided, all are set to 1.
    '''

    # Three qubits because for 2 we get H_O = 0
    assert n_qubits >= 3

    if coeff == None:
        'Setting default values for the coefficients'
        coeff = [[1.0, 1.0, 1.0, 1.0] for i in range(n_qubits)]

    # Even terms of the Hamiltonian
    # (summing over individual pair-wise elements)
    H_E = sum((get_hamiltonian_n_site_terms(i, coeff[i], n_qubits)
               for i in range(0, n_qubits-1, 2)))

    # Odd terms of the Hamiltonian
    # (summing over individual pair-wise elements)
    H_O = sum((get_hamiltonian_n_site_terms(i, coeff[i], n_qubits)
               for i in range(1, n_qubits-1, 2)))

    # adding final Z term at the Nth site
    final_term = SparsePauliOp("I" * (n_qubits - 1) + "Z")
    final_term *= coeff[n_qubits-1][3]
    if (n_qubits % 2) == 0:
        H_E += final_term     
    else:
        H_O += final_term     

    # Returns the list of the two sets of terms
    return [H_E, H_O]


if __name__ == '__main__':
    num_q = 3
    # XX YY ZZ, Z
    ham_coeffs = ([[0.75/2, 0.75/2, 0.0, 0.65]]
                  + [[0.5, 0.5, 0.0, 1.0]
                  for i in range(num_q-1)])

    spin_chain_hamiltonian = get_heisenberg_hamiltonian(num_q,
                                                        ham_coeffs)
    print('Hamiltonian separated into even and odd components:')
    print(spin_chain_hamiltonian)
    print('Hamiltonian combining even and odd components:')
    print(sum(spin_chain_hamiltonian))



import numpy as np
import functools
import sys

from mpsqd.utils import MPS
from mpsqd.utils.split import split_rq,split_qr

from mpsqd.tdvp import ttfunc as ttf


"""
Modified from Mpsqd (https://github.com/qiangshi-group/MPSQD)

The MIT License (MIT)
Copyright (c) 2020 mpsqd team, Qiang Shi group

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

By Weizhong Guan; Peng Bao; Jiawei Peng; Zhenggang Lan and Qiang Shi 
J. Chem. Phys. 161, 122501 (2024)
Changes:
    add rk4slices as argument (for update_type='rk4')
"""
def tdvp1site(rin,pall,dt,update_type='krylov',mmax=30,nsteps=1,rk4slices = 10):
  argdict = {'dt':dt,'mmax':mmax,'nsteps':nsteps,'update_type':update_type, 'rk4slices':rk4slices}
  try:
    for key in ['nsteps','mmax']:
      argdict[key] = int(argdict[key])
    for key in ['dt']:
      argdict[key] = float(argdict[key])
    for key in ['dt','nsteps','mmax']:
      if (argdict[key] <= 0):
        print("Wrong value in tdvp: ",key," = ",argdict[key])
        sys.exit()
    key = 'update_type'
    argdict[key] = argdict[key].lower()
    if (not argdict[key] in ['krylov','rk4']):
      print("Wrong value in tdvp: update_type = ",update_type)
      sys.exit()
    if(rin.__class__.__name__ != "MPS"):
      print("Wrong value in tdvp: improper MPS")
      sys.exit()
    if(pall.__class__.__name__ != "MPO"):
      print("Wrong value in tdvp: improper MPO")
      sys.exit()
  except (ValueError):
    print("Wrong value in tdvp: ",key," = ",argdict[key])
    sys.exit()
  for istep in range(nsteps):
    rin = _tdvp1(rin,pall,argdict['dt'],argdict['update_type'],argdict['mmax'],argdict['rk4slices'])
  return rin


"""
Modified from Mpsqd (https://github.com/qiangshi-group/MPSQD)

The MIT License (MIT)
Copyright (c) 2020 mpsqd team, Qiang Shi group

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

By Weizhong Guan; Peng Bao; Jiawei Peng; Zhenggang Lan and Qiang Shi 
J. Chem. Phys. 161, 122501 (2024)
Changes:
    add rk4slices as argument (for update_type='rk4')
"""
def _tdvp1(rin,pall,dt,update_type='krylov',mmax=30, rk4slices = 10):

  dt2 = 0.5*dt
  if(update_type=='krylov'):
    update_v = functools.partial(ttf.expmv,mmax=mmax,dt=dt2)
  else:
    update_v = functools.partial(ttf.update_rk4,dt=dt2,rk4slices=rk4slices)

  nlen = len(rin.nodes)
  r1 = MPS(nlen,rin.nb)
  r2 = MPS(nlen,rin.nb)
  r3 = MPS(nlen,rin.nb)

  phia = []

  # phia[nlevel+2]
  vtmp = np.ones((1,1,1),dtype=np.complex128)
  phia.append(vtmp)

#======================================================
# ortho from right
  r, q = split_rq(rin.nodes[nlen-1])
  
  r1.nodes.append(q)
  u1 = r
# phia[nlevel+1]
  phia.append(ttf.phia_next(phia[0],pall.nodes[nlen-1],q,q,1))
#-------------------------------------------------------
# intermediate terms
  for i in range(nlen-2,0,-1):

    rtmp = np.tensordot(rin.nodes[i], u1, axes=((2),(0)))

    r, q = split_rq(rtmp)
    r1.nodes.append(q)
    u1 = r

    phia.append(ttf.phia_next(phia[nlen-1-i],pall.nodes[i],q,q,1))

# the left matrix

  rtmp = np.tensordot(rin.nodes[0], u1, axes=((2),(0)))
  r1.nodes.append(rtmp)

  r1.nodes.reverse()

# add phia[0] and reverse to normal
  vtmp = np.ones((1,1,1),dtype=np.complex128)
  phia.append(vtmp)
  phia.reverse()

#===============================================
###### the first part of KSL, from left to right
  for i in range(nlen-1):
    phi1 = phia[i]
    phi2 = phia[i+1]

    if (i == 0):
      ksol = r1.nodes[i].copy()
    else:
      ksol = r2.nodes[i].copy()


    #for islice in range(pa.rk4slice):
    ksol = update_v(yy=ksol, phi1=phi1, phi2=phi2, mat1=pall.nodes[i])
    q, r = split_qr(ksol)


    if (i == 0):
      r2.nodes.append(q)
    else:
      r2.nodes[i] = q


    phi1 = ttf.phia_next(phi1,pall.nodes[i],q,q,0)
    phia[i+1] = phi1

# ?? need to copy
    ssol = r
    ssol = update_v(yy=ssol, phi1=phi1, phi2=phi2)

    rtmp1 = np.tensordot(ssol,r1.nodes[i+1],axes=((1),(0)))
    r2.nodes.append(rtmp1)

#--------------------------------------------------------
### right most part

  phi1 = phia[nlen-1]
  phi2 = phia[nlen]
  ksol = r2.nodes[nlen-1].copy()
  ksol = update_v(yy=ksol, phi1=phi1, phi2=phi2, mat1=pall.nodes[nlen-1])

  r2.nodes[nlen-1] = ksol
#===================================================================
###### the second part of KSL, from right to left
  for i in range(nlen-1,0,-1):

    phi1 = phia[i]
    phi2 = phia[i+1]

    if (i==nlen-1):
      ksol = r2.nodes[i].copy()
    else:
      ksol = r3.nodes[nlen-1-i].copy()

    #for islice in range(pa.rk4slice):
    ksol = update_v(yy=ksol, phi1=phi1, phi2=phi2, mat1=pall.nodes[i])
    r, q = split_rq(ksol)

    if (i==nlen-1):
      r3.nodes.append(q)
    else:
      r3.nodes[nlen-1-i] = q

    phi2 = ttf.phia_next(phi2,pall.nodes[i],q,q,1)
    phia[i] = phi2

    ssol = r

    #for islice in range(pa.rk4slice):
    ssol = update_v(yy=ssol, phi1=phi1, phi2=phi2)

    rtmp1 = np.tensordot(r2.nodes[i-1],ssol,axes=((2),(0)))
    r3.nodes.append(rtmp1)

  r3.nodes.reverse()
  ### the left most matrix
  phi1 = phia[0]
  phi2 = phia[1]
  ksol = r3.nodes[0].copy()
  ksol = update_v(yy=ksol, phi1=phi1, phi2=phi2, mat1=pall.nodes[0])

  r3.nodes[0] = ksol

  return r3


import numpy as np
import scipy.fft as sfft

#the module in the current package
from . import params as pa
    
def x2k_wave(dx,psi):
    """
    transform the wavefunction from x space to k space
    dx: the interval of the x-space grid point
    psi: the wavefunction in the x-space
    """
    pre_fac = dx/(2*np.pi)**0.5
    psik = sfft.fft(psi)*pre_fac
    return psik

def k2x_wave(dx,psik):
    """
    transform the wavefunction from x space to k space
    dx: the interval of the x-space grid point
    psi: the wavefunction in the k-space
    """
    pre_fac = (2*np.pi)**0.5/dx
    psix  = sfft.ifft(psik.copy())*pre_fac
    return psix

def nested_kronecker_product(pauli_str):
    '''
    Handles Kronecker Products for list (i.e., pauli_str = 'ZZZ' will evaluate Z Z Z).
    Given string 'pauli_str' this evaluates the kronecker product of all elements.
    '''

    # Define a dictionary with the four Pauli matrices:
    pms = {'I': pa.I,'X': pa.X,'Y': pa.Y,'Z': pa.Z}

    result = 1
    for i in range(len(pauli_str)):
     result = np.kron(result,pms[pauli_str[i]])
    return result


def pauli_to_ham(pauli_dict, Nqb):
    ''' 
        Function that Assembles the Hamiltonian based on their Pauli string
        
        pauli_dict: A dictionary that contain all the pauli string and the value of the Hamiltonian
            (the key is pauli string and the value is coefficient)
        
        Nqb: number of qubits, should match the length of the pauli string
        
        return: Hamiltonian matrix 
    '''
        
    Hmat = np.zeros((2**Nqb,2**Nqb),dtype=np.complex128)
    for key in pauli_dict:
        Hmat += pauli_dict[key]*nested_kronecker_product(key)

    return Hmat

def ham_to_pauli(Ham_arr, Nqb, tol=1E-5):
    '''
    Function that decomposes `Ham_arr` into a sum of Pauli strings.
    result: a dictionary with the key is pauli string and the value is coefficient
    '''
    import itertools

    pauli_keys = ['I','X','Y','Z'] # Keys of the dictionary

    if(2**Nqb != Ham_arr.shape[0]):
        print('Nqb and Matrix size not matched!')

    # Make all possible tensor products of Pauli matrices sigma
    sigma_combinations = list(itertools.product(pauli_keys, repeat=Nqb))

    result = {} # Initialize an empty dictionary to the results
    for ii in range(len(sigma_combinations)):
        pauli_str = ''.join(sigma_combinations[ii])

        # Evaluate the Kronecker product of the matrix array
        tmp_p_matrix = nested_kronecker_product(pauli_str)

        # Compute the coefficient for each Pauli string
        a_coeff = (1/(2**Nqb)) * np.trace(tmp_p_matrix @ Ham_arr)

        # If the coefficient is non-zero, we want to use it!
        if abs(a_coeff) > tol:
            result[pauli_str] = a_coeff.real

    return result

def trans_basis(operator,nbasis,psi_newbasis):
    """
    operator: the operator matrix in the old basis
    nbasis: the truncation in the new basis
    psi_newbasis: the new basis wave function expressed in the old basis 
               psi_newbasis[i,j] = <i|psi_newbasis|j>, with i is old basis, j new basis
    """
    
    operator_new = np.zeros((nbasis,nbasis),dtype=np.complex128)
    
    for i in range(nbasis):
      for j in range(nbasis):
        operator_new[i,j] = np.dot(np.dot(psi_newbasis[:,i].conj(),operator),psi_newbasis[:,j])

    return operator_new

def trans_basis_diag(diag_opr,nbasis,psi_newbasis):
    """
    diag_opr: the array represent the diagnoal operator in the old basis
    nbasis: the truncation in the new basis
    psi_newbasis: the new basis wave function expressed in the old basis 
               psi_newbasis[i,j] = <i|psi_newbasis|j>, with i is old basis, j new basis
    """
    
    operator_new = np.zeros((nbasis,nbasis),dtype=np.complex128)
    
    for i in range(nbasis):
      for j in range(nbasis):
        operator_new[i,j] = np.dot(np.multiply(psi_newbasis[:,i].conj(),diag_opr),psi_newbasis[:,j])

    return operator_new


from pathlib import Path


def load_dataset(ds_path: Path) -> None:
    """Loads a dataset from the specified path.

    This function is intended to load a dataset from the given file path.
    Currently, it is not implemented and always raises a NotImplementedError.

    Args:
        ds_path: The path to the dataset file.

    Raises:
        NotImplementedError: Always raised as the function is not yet implemented.
    """
    raise NotImplementedError("Not implemented")



import qflux

from qflux.closed_systems import DynamicsCS
print("Checking Closed Systems...")
print(dir(DynamicsCS))

from qflux.variational_methods import qmad
print("Checking Variational Methods...")
print(dir(qmad))

print("Checking qmad modules...")

from qflux.variational_methods.qmad import effh, ansatz, solver
print("Methods from effh:")
print(dir(effh))

print("Methods from ansatz:")
print(dir(ansatz))

print("Methods from solver:")
print(dir(solver))
